/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Drexel University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Joe Kopena <tjkopena@cs.drexel.edu>
 */

#include <iostream>
#include <fstream>
#include <vector>
#include <sstream>

#include "net-debug.h"

#include "tokenizer.h"
#include "node-style.h"
#include "node-style-driver.h"

using namespace ns3;

uint8_t
NodeStyleDriver::HexToDec(std::string v)
{
  uint8_t val;

  // NET_LOG(v);

  if (v.length() != 2) {
    NET_ERR("Did not understand double digit hex value " << v);
  }

  const char *str = v.c_str();

  if (str[0] >= '0' && str[0] <= '9') {
    val = (str[0] - '0') * 16;
  } else if (str[0] >= 'a' && str[0] <= 'f') {
    val = ((str[0] - 'a')+10) * 16;
  } else if (str[0] >= 'A' && str[0] <= 'F') {
    val = ((str[0] - 'A')+10) * 16;
  } else {
    NET_ERR("Did not understand double digit hex value " << v);
  }

  // NET_LOG((int) val);

  if (str[1] >= '0' && str[1] <= '9') {
    val += (str[1] - '0');
  } else if (str[1] >= 'a' && str[1] <= 'f') {
    val += (str[1] - 'a')+10;
  } else if (str[1] >= 'A' && str[1] <= 'F') {
    val += (str[1] - 'A')+10;
  } else {
    NET_ERR("Did not understand double digit hex value " << v);
  }

  // NET_LOG((int) val);

  return val;

  // end NodeStyleDriver::HexToDec
}

bool
NodeStyleDriver::LoadNodeStyles(NodeStyleMap &nodeStyles,
				std::string filename)
{

  NET_STATE(filename);

  char str[BUFFER_LEN];
  std::vector<std::string> tokens;

  std::ifstream file(filename.c_str());
  if (file.is_open()) {

    uint32_t linecount = 0;

    while(!file.eof()) {

      file.getline(str, BUFFER_LEN);
      linecount++;

      NET_LOG(str);

      Tokenizer::Tokenize(tokens, str, BUFFER_LEN);

      if (tokens.size() >= 1) {

	std::vector<std::string>::iterator i = tokens.begin();
	std::string styleKey(*i);
	i++;

	NodeStyle style;

	while (i != tokens.end()) {
	  std::string command(*i);

	  if (command == "linewidth") {
	    if (++i == tokens.end()) {
	      NET_ERR("Linewidth command in " << filename << " line " <<
		      linecount << " but no parameter.");
	    } else {
	      std::istringstream val(*i);
	      if (!(val >> style.m_linewidth)) {
		NET_ERR("Linewidth command in " << filename << " line " <<
			linecount << " had unparsable parameter.");
	      } else {
		style.m_hasLinewidth = true;
		NET_LOG("Style " << styleKey << " line width " <<
			style.m_linewidth);
	      }
	      // end had a parameter
	    }

	    // end command linewidth
	  } else if (command == "color") {
	    if (++i == tokens.end()) {
	      NET_ERR("color command in " << filename << " line " <<
		      linecount << " but no parameter.");
	    } else {
	      std::string color(*i);
	      if (color.length() != 6 && color.length() != 8) {
		NET_ERR("color command parameter in " << filename <<
			", line " << linecount << " wrong length.");
	      } else if (color.find_first_not_of("0123456789abcdefABCDEF")
			 != std::string::npos) {
		NET_ERR("color command parameter in " << filename <<
			", line " << linecount << " unparsable.");
	      } else {

		uint8_t red = HexToDec(color.substr(0,2));
		uint8_t green = HexToDec(color.substr(2,2));
		uint8_t blue = HexToDec(color.substr(4,2));
		uint8_t alpha = 255;

		if (color.length() == 8) {
		  alpha = HexToDec(color.substr(6,2));
		}

		NET_LOG("Color red " << (int) red <<
			" green " << (int) green <<
			" blue " << (int) blue <<
			" alpha " << (int) alpha);

		style.m_hasColor = true;
		style.m_r = (double) red/255.0;
		style.m_g = (double) green/255.0;
		style.m_b = (double) blue/255.0;
		style.m_a = (double) alpha/255.0;

		// end valid color parameter
	      }

	      // end had a parameter
	    }

	    // end color command
	  } else if (command == "linecolor") {
	    if (++i == tokens.end()) {
	      NET_ERR("linecolor command in " << filename << " line " <<
		      linecount << " but no parameter.");
	    } else {
	      std::string color(*i);
	      if (color.length() != 6 && color.length() != 8) {
		NET_ERR("linecolor command parameter in " << filename <<
			", line " << linecount << " wrong length.");
	      } else if (color.find_first_not_of("0123456789abcdefABCDEF")
			 != std::string::npos) {
		NET_ERR("linecolor command parameter in " << filename <<
			", line " << linecount << " unparsable.");
	      } else {

		uint8_t red = HexToDec(color.substr(0,2));
		uint8_t green = HexToDec(color.substr(2,2));
		uint8_t blue = HexToDec(color.substr(4,2));
		uint8_t alpha = 255;

		if (color.length() == 8) {
		  alpha = HexToDec(color.substr(6,2));
		}

		NET_LOG("Line color red " << (int) red <<
			" green " << (int) green <<
			" blue " << (int) blue <<
			" alpha " << (int) alpha);

		style.m_hasLineColor = true;
		style.m_lr = (double) red/255.0;
		style.m_lg = (double) green/255.0;
		style.m_lb = (double) blue/255.0;
		style.m_la = (double) alpha/255.0;

		// end valid color parameter
	      }

	      // end had a parameter
	    }

	    // end linecolor command
	  } else if (command == "radius") {
	    if (++i == tokens.end()) {
	      NET_ERR("Radius command in " << filename << " line " <<
		      linecount << " but no parameter.");
	    } else {
	      std::istringstream val(*i);
	      if (!(val >> style.m_radius)) {
		NET_ERR("Radius command in " << filename << " line " <<
			linecount << " had unparsable parameter.");
	      } else {
		style.m_hasRadius = true;
		NET_LOG("Style " << styleKey << " radius " <<
			style.m_radius);
	      }
	      // end had a parameter
	    }

	    // end radius command
          } else if (command == "shape") {
	    if (++i == tokens.end()) {
	      NET_ERR("Radius command in " << filename << " line " <<
		      linecount << " but no parameter.");
	    } else {
              std::string type = *i;
              if (type == "square") {
                style.m_hasShape = true;
                style.m_shape = NODE_SQUARE;
                NET_LOG("Set " << styleKey << " to square.");
              } else if (type == "circle") {
                style.m_hasShape = true;
                style.m_shape = NODE_CIRCLE;
                NET_LOG("Set " << styleKey << " to circle.");
              } else {
                NET_ERR("Unknown shape parameter " << type <<
                        " in file " << filename << ", line " << linecount);
              }

              // has a type
            }
            // end shape command
          } else if (command == "double") {
            style.m_hasDouble = true;
            style.m_double = true;
            NET_LOG("Set " << styleKey << " to double border.");
            // end double command
	  } else {
	    NET_ERR("Unknown style command " << command << " in " <<
		    filename << ", line " << linecount);
	  }

	  i++;
	  // end looping over tokens
	}

	nodeStyles[styleKey] = style;

	// end enough tokens
      }

      tokens.clear();

      // end looping over file until eof
    }

    NET_LOG("Read " << linecount << " lines.");

    file.close();

    // end file is open
  } else {
    NET_ERR("Could not open node style file " << filename);
    return true;
  }

  return false;

  // end NodeStyleDriver::LoadNodeStyles
}
