/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Drexel University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Joe Kopena <tjkopena@cs.drexel.edu>
 */

#include "net-debug.h"

#include "net-world-model.h"

using namespace ns3;

NetWorldModel::NetWorldModel() :
  m_leftx(0),
  m_bottomy(0),
  m_rightx(400),
  m_topy(400)
{
  NET_LOG_NOARGS();

  // end NetWorldModel::NetWorldModel
}
NetWorldModel::~NetWorldModel()
{

  ClearNodes();

  uint32_t c = 0;
  for (NetWorldEventVector::iterator i = m_events.begin();
       i != m_events.end(); i++) {
    delete *i;
    c++;
  }
  NET_LOG("Deleted " << c << " events");
  m_events.clear();

  // end NetWorldModel::~NetWorldModel
}

void
NetWorldModel::ClearNodes()
{

  uint32_t c = 0;
  for (NodeIDMap::iterator i = m_nodes.begin(); i != m_nodes.end(); i++) {
    NetWorldNode *n = (*i).second;
    delete n;
    c++;
  }
  NET_LOG("Deleted " << c << " nodes");
  m_nodes.clear();
  m_ips.clear();

  // end NetWorldModel::Clear
}

void
NetWorldModel::SetArenaBounds(double x1, double y1,
			      double x2, double y2)
{

  if (x1 <= x2) {
    m_leftx = x1;
    m_rightx = x2;
  } else {
    m_leftx = x2;
    m_rightx = x1;
  }

  if (y1 <= y2) {
    m_bottomy = y1;
    m_topy = y2;
  } else {
    m_bottomy = y2;
    m_topy = y1;
  }

  NET_LOG("bounds " << m_leftx << ", " << m_bottomy <<
	  " --- " << m_rightx << ", " << m_topy);

  // end NetWorldModel::SetArenaDimensions
}

double 
NetWorldModel::GetLeft()
{
  return m_leftx;
}
double
NetWorldModel::GetRight()
{
  return m_rightx;
}
double
NetWorldModel::GetTop()
{
  return m_topy;
}
double
NetWorldModel::GetBottom()
{
  return m_bottomy;
}

NodeIDMap::iterator
NetWorldModel::NodesBegin()
{
  return m_nodes.begin();
  // end NetWorldModel::NodesBegin
}
NodeIDMap::iterator
NetWorldModel::NodesEnd()
{
  return m_nodes.end();
  // end NetWorldModel::NodesEnd
}

void
NetWorldModel::PushEvent(NetWorldEvent *e)
{

  m_events.push_back(e);

  // end NetWorldModel::PushEvent
}

//----------------------------------
void
NetWorldModel::Start()
{
  ClearNodes();
  m_currentEvent = m_events.begin();
  // end NetWorldModel::Start
}
bool
NetWorldModel::End()
{
  return (m_currentEvent == m_events.end());

  // end NetWorldModel::Start
}
void
NetWorldModel::ProcessUpTo(NetWorldTime t)
{

  // NET_LOG("Bound " << t);

  while ((m_currentEvent != m_events.end()) &&
         (*m_currentEvent)->GetTimestamp() <= t) {
    (*m_currentEvent)->Process((NetWorldModel *) this);
    m_currentEvent++;
  }

  // end NetWorldModel::ProcessUpTo
}

NetWorldTime
NetWorldModel::GetLastTimestamp()
{
  if (m_events.size() <= 0) {
    return 0;
  } else {
    return m_events.back()->GetTimestamp();
  }

  // end NetWorldModel::GetLastTimestamp
}


NetWorldNode *
NetWorldModel::GetNode(uint32_t id)
{
  NodeIDMap::iterator i = m_nodes.find(id);

  if (i == m_nodes.end()) {
    NetWorldNode *n = new NetWorldNode(id);
    if (n != 0) {
      m_nodes[id] = n;
      // NET_LOG("New node " << id);
      return n;
    } else {
      NET_ERR("No memory for new node " << id);
      return 0; // Should throw exception
    }

  } else {
    // NET_LOG("Retrieve node " << id);
    return (*i).second;
  }

  // end NetWorldModel::GetNode
}
NetWorldNode *
NetWorldModel::GetNodeByIP(std::string ip)
{
  NodeIPMap::iterator i = m_ips.find(ip);
  if (i == m_ips.end())
    return 0;

  return (*i).second;

  // end NetWorldModel::GetNode
}

//----------------------------------
void
NetWorldModel::MapIP(std::string ip, uint32_t nodeId)
{
  NET_LOG("Map " << ip << " -> " << nodeId);

  NetWorldNode *node = GetNode(nodeId);
  if (node == 0) {
    NET_ERR("Could not get or create node to map.");
    return;
  }

  if (m_ips.find(ip) != m_ips.end()) {
    NET_LOG("Warning: IP " << ip << " already points to node " <<
            m_ips[ip]->GetID());
  }

  node->AddIP(ip);
  m_ips[ip] = node;

  // end NetWorldModel::MapIP
}
void
NetWorldModel::SetPosition(uint32_t nodeId, NetWorldTime time,
                           double x, double y, double z,
                           double xv, double yv, double zv)
{

  // NET_LOG("Course change for " << nodeId);

  NetWorldNode *node = GetNode(nodeId);
  if (node == 0) {
    NET_ERR("Could not get or create node to map.");
    return;
  }

  node->SetPosition(x, y, time, xv, yv);

  // end NetWorldModel::SetPosition
}
void
NetWorldModel::AddLinkState(uint32_t nodeID, std::string target,
                            std::string state)
{

  NET_LOG("Add link state " << state << " to " << nodeID << " -> " << target);

  NetWorldNode *node = GetNode(nodeID);
  if (node == 0) {
    NET_ERR("Could not get or create node to map.");
    return;
  }

  node->AddLinkState(target, state);

  // end NetWorldModel::AddLinkState
}
void
NetWorldModel::RemoveLinkState(uint32_t nodeID, std::string target,
                               std::string state)
{

  NET_LOG("Remove link state " << state << " from " <<nodeID << " -> " <<target);

  NetWorldNode *node = GetNode(nodeID);
  if (node == 0) {
    NET_ERR("Could not get or create node to map.");
    return;
  }

  node->RemoveLinkState(target, state);

  // end NetWorldModel::RemoveLinkState
}

void
NetWorldModel::AddNodeState(uint32_t nodeID, std::string state)
{
  NET_LOG("Add state " << state << " from " << nodeID);

  NetWorldNode *node = GetNode(nodeID);
  if (node == 0) {
    NET_ERR("Could not get or create node " << nodeID <<
            " on which to set state.");
    return;
  }

  node->AddNodeState(state);

  // end NetWorldModel::AddNodeState
}

void
NetWorldModel::RemoveNodeState(uint32_t nodeID, std::string state)
{
  NET_LOG("Remove state " << state << " from " << nodeID);

  NetWorldNode *node = GetNode(nodeID);
  if (node == 0) {
    NET_ERR("Could not get or create node " << nodeID <<
            " on which to set state.");
    return;
  }

  node->RemoveNodeState(state);
  // end NetWorldModel::RemoveNodeState
}


//----------------------------------------------------------------------
//----------------------------------------------------------------------
NetWorldEvent::NetWorldEvent(NetWorldTime t) :
  m_timestamp(t)
{
  // end NetWorldEvent::NetWorldEvent
}
NetWorldEvent::~NetWorldEvent()
{
  // end NetWorldEvent::NetWorldEvent
}

NetWorldTime
NetWorldEvent::GetTimestamp()
{
  return m_timestamp;

  // end NetWorldEvent::GetTimestamp
}

//----------------------------------------------
NetWorldIPEvent::NetWorldIPEvent(NetWorldTime t,
                                 uint32_t id, std::string ip) :
  NetWorldEvent(t),
  m_node(id),
  m_ip(ip)
{

  // end NetWorldIPEvent::NetWorldIPEvent
}

void
NetWorldIPEvent::Process(NetWorldModel *worldModel)
{

  NET_LOG("Bind ip for " << m_node << " to " << m_ip);

  worldModel->MapIP(m_ip, m_node);

  // end NetWorldIPEvent::Process
}

//----------------------------------------------
NetWorldPositionEvent::NetWorldPositionEvent(NetWorldTime t, uint32_t id,
                                             double x, double y, double z,
                                             double xv, double yv, double zv) :
  NetWorldEvent(t),
  m_node(id),
  m_x(x),
  m_y(y),
  m_z(z),
  m_xvel(xv),
  m_yvel(yv),
  m_zvel(zv)
{

  // end NetWorldPositionEvent::NetWorldPositionEvent
}

void
NetWorldPositionEvent::Process(NetWorldModel *worldModel)
{

  /*
  NET_LOG("Course change for " << m_node << " at t" << GetTimestamp() << ", "<<
          m_x << ", " << m_y << ", " << m_z << " to " <<
          m_xvel << ", " << m_yvel << ", " << m_zvel);
  */

  worldModel->SetPosition(m_node, GetTimestamp(),
                          m_x, m_y, m_z, m_xvel, m_yvel, m_zvel);

  // end NetWorldPositionEvent::Process
}

//----------------------------------------------
NetWorldLinkStateEvent::NetWorldLinkStateEvent(NetWorldTime t,
                                               uint32_t id, std::string target,
                                               bool link, std::string state) :
  NetWorldEvent(t),
  m_node(id),
  m_target(target),
  m_link(link),
  m_state(state)
{
  NET_LOG(id << " -> " << target << " state " << state);

  // end NetWorldLinkStateEvent::NetWorldLinkStateEvent
}

void
NetWorldLinkStateEvent::Process(NetWorldModel *worldModel)
{

  if (m_link) {
    NET_LOG("Add state " << m_state << " to link " <<
            m_node << " -> " << m_target);
    worldModel->AddLinkState(m_node, m_target, m_state);
  } else {
    NET_LOG("Remove state " << m_state << " to link " <<
            m_node << " -> " << m_target);
    worldModel->RemoveLinkState(m_node, m_target, m_state);
  }

  // end NetWorldLinkStateEvent::Process
}

//----------------------------------------------
NetWorldNodeStateEvent::NetWorldNodeStateEvent(NetWorldTime t,
                                               uint32_t id,
                                               bool add, std::string state) :
  NetWorldEvent(t),
  m_node(id),
  m_add(add),
  m_state(state)
{
  NET_LOG(id << " state " << state);
  // end NetWorldNodeStateEvent::NetWorldNodeStateEvent
}

void
NetWorldNodeStateEvent::Process(NetWorldModel *worldModel)
{

  if (m_add) {
    NET_LOG("Add state " << m_state <<  " to " << m_node);
    worldModel->AddNodeState(m_node, m_state);
  } else {
    NET_LOG("Remove state " << m_state <<  " to " << m_node);
    worldModel->RemoveNodeState(m_node, m_state);
  }

  // end NetWorldStateEvent::Process
}
