/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Drexel University
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Joe Kopena <tjkopena@cs.drexel.edu>
 */

#include <cairomm/context.h>


#include "net-debug.h"
#include "net-view-area.h"

using namespace ns3;

NetViewArea::NetViewArea() :
  m_worldModel(0),
  m_viewTime(0)
{

  set_size_request(400, 400);

  // end NetViewArea::NetViewArea
}

NetViewArea::~NetViewArea()
{
  // end NetViewArea::~NetViewArea
}


//----------------------------------------------------------------------
//------------------------------------------------------
void
NetViewArea::SetWorldModel(NetWorldModel *worldModel)
{
  NET_LOG_NOARGS();

  // assert(worldModel != 0);
  m_worldModel = worldModel;
  queue_draw();
  // end NetViewArea::SetWorldModel
}

void
NetViewArea::SetAnimationTime(NetWorldTime time)
{
  // NET_LOG("Current t " << m_viewTime << ", new t " << time);
  m_viewTime = time;
  queue_draw();
  // end NetViewArea::SetAnimationTime
}

void
NetViewArea::SetLinkStyleMap(LinkStyleMap styles)
{
  m_linkStyles = styles;

  /*
  NET_LOG("Link styles:");
  for (ns3::LinkStyleMap::iterator i = m_linkStyles.begin();
       i != m_linkStyles.end(); i++) {
    NET_LOG("Found style " << ((*i).first));
  }
  */

  // end NetViewArea::SetLinkStyleMap
}

void
NetViewArea::SetNodeStyleMap(NodeStyleMap styles)
{
  m_nodeStyles = styles;
  // end NetViewArea::SetNodeStyleMap
}

//----------------------------------------------------------------------
//------------------------------------------------------
bool NetViewArea::on_expose_event(GdkEventExpose* event)
{

  // NET_LOG_NOARGS();

  Glib::RefPtr<Gdk::Window> window = get_window();
  if(window != 0 && m_worldModel != 0) {

    Gtk::Allocation allocation = get_allocation();

    const double minx = m_worldModel->GetLeft();
    const double miny = m_worldModel->GetBottom();
    const double maxx = m_worldModel->GetRight();
    const double maxy = m_worldModel->GetTop();

    const int width = allocation.get_width();
    const int height = allocation.get_height();

    double xscale = width/(maxx-minx);
    double yscale = height/(maxy-miny);

    /*
    NET_LOG("Origin " << minx << ", " << miny << "; scale " <<
	    xscale << ", " << yscale);
    */

    Cairo::RefPtr<Cairo::Context> cr = window->create_cairo_context();

    cr->rectangle(event->area.x, event->area.y,
                  event->area.width, event->area.height);
    cr->clip();

    bool hasPos;
    double x, y, tx, ty;
    NetWorldNode *n;

    //-- Draw links ----------------------------------------------------
    for (NodeIDMap::iterator i = m_worldModel->NodesBegin();
	 i != m_worldModel->NodesEnd(); i++) {
      n = (*i).second;
      n->GetPosition(hasPos, x, y, m_viewTime);
      if (hasPos) {

        /*
        NET_LOG("Drawing node links " << (*i).first <<
                " at " << x << ", " << y);
        */

        x -= minx;
        y -= miny;

        x *= xscale;
        y *= yscale;

        for (LinkMap::iterator l = n->LinksBegin();
             l != n->LinksEnd(); l++) {
          std::string ip = (*l).first;
          LinkStateSet linkStates = (*l).second;

          if (linkStates.begin() != linkStates.end()) {
            LinkStyle style;
            for (LinkStateSet::iterator i = linkStates.begin();
                 i != linkStates.end();
                 i++) {

              std::string styleKey = *i;
              if (m_linkStyles.find(styleKey) != m_linkStyles.end()) {
                // NET_LOG("Found link style " << styleKey << " in library.");

                LinkStyle x = m_linkStyles[styleKey];
                if (x.m_hasLinewidth) {
                  style.m_linewidth = x.m_linewidth;
                }
                if (x.m_hasColor) {
                  style.m_r = x.m_r;
                  style.m_g = x.m_g;
                  style.m_b = x.m_b;
                  style.m_a = x.m_a;
                }
                // end found the style
              }

                // end iterating over styles for this link
            }

            NetWorldNode *target = m_worldModel->GetNodeByIP(ip);
            if (target == 0) {
              NET_LOG("No node for target ip for link " <<
                      n->GetID() << " -> " << ip);
              continue;
            }

            target->GetPosition(hasPos, tx, ty, m_viewTime);
            if (hasPos) {

              /*
              NET_LOG("Drawing link " << n->GetID() << " -> " <<
                      target->GetID() << " (" << ip << ") @ " <<
                      x << ", " << y << " -> " << tx << ", " << ty);
              */

              tx -= minx;
              ty -= miny;

              tx *= xscale;
              ty *= yscale;

              cr->save();

              cr->set_line_width(style.m_linewidth);
              cr->set_source_rgba(style.m_r, style.m_g, style.m_b, style.m_a);

              cr->move_to(x, height-y);
              cr->line_to(tx, height-ty);
              cr->stroke();
              cr->restore();

              // end target has position
            }

          // end found link styles
          }
          // end looping over links
        }

        // end node had a position
      }
      // end looping over nodes
    }

    //-- Draw nodes ----------------------------------------------------
    for (NodeIDMap::iterator i = m_worldModel->NodesBegin();
	 i != m_worldModel->NodesEnd(); i++) {
      n = (*i).second;
      n->GetPosition(hasPos, x, y, m_viewTime);
      if (hasPos) {
        // NET_LOG("Drawing node " << (*i).first << " at " << x << ", " << y);

        NodeStyle nodeStyle;
        for (NodeStateSet::iterator nssi = n->StatesBegin();
             nssi != n->StatesEnd(); nssi++) {
          std::string state = *nssi;
          // NET_LOG("State " << state << " for " << n->GetID());

          if (m_nodeStyles.find(state) != m_nodeStyles.end()) {
            // NET_LOG("Found node style " << state << " in library.");

            NodeStyle ns = m_nodeStyles[state];
            if (ns.m_hasLinewidth)
              nodeStyle.m_linewidth = ns.m_linewidth;
            if (ns.m_hasLineColor) {
              nodeStyle.m_lr = ns.m_lr;
              nodeStyle.m_lg = ns.m_lg;
              nodeStyle.m_lb = ns.m_lb;
              nodeStyle.m_la = ns.m_la;
            }
            if (ns.m_hasColor) {
              nodeStyle.m_r = ns.m_r;
              nodeStyle.m_g = ns.m_g;
              nodeStyle.m_b = ns.m_b;
              nodeStyle.m_a = ns.m_a;
            }
            if (ns.m_hasRadius) {
              nodeStyle.m_radius = ns.m_radius;
            }
            if (ns.m_hasShape) {
              nodeStyle.m_shape = ns.m_shape;
            }
            if (ns.m_hasDouble) {
              nodeStyle.m_double = ns.m_double;
            }
            // end found a style for state
          }

              // end looping over node states
        }

        x -= minx;
        y -= miny;

        x *= xscale;
        y *= yscale;

        cr->save();
        switch (nodeStyle.m_shape) {
        case NODE_SQUARE:
          cr->move_to(x-nodeStyle.m_radius, (height-y)-nodeStyle.m_radius);
          cr->line_to(x-nodeStyle.m_radius, (height-y)+nodeStyle.m_radius);
          cr->line_to(x+nodeStyle.m_radius, (height-y)+nodeStyle.m_radius);
          cr->line_to(x+nodeStyle.m_radius, (height-y)-nodeStyle.m_radius);
          cr->close_path();
          break;

        case NODE_CIRCLE:
        default:
          cr->arc(x, height-y, nodeStyle.m_radius, 0, 2 * M_PI);
          break;
          // end switching node shape
        }
        cr->set_source_rgba(nodeStyle.m_r, nodeStyle.m_g, nodeStyle.m_b,
                            nodeStyle.m_a);
        cr->set_line_width(nodeStyle.m_linewidth);
        cr->fill_preserve();
        cr->set_source_rgba(nodeStyle.m_lr, nodeStyle.m_lg, nodeStyle.m_lb,
                            nodeStyle.m_la);
        cr->stroke();

        if (nodeStyle.m_double) {
          switch (nodeStyle.m_shape) {
          case NODE_SQUARE:
            cr->move_to(x-nodeStyle.m_radius-nodeStyle.m_linewidth*2,
                        (height-y)-nodeStyle.m_radius-nodeStyle.m_linewidth*2);
            cr->line_to(x-nodeStyle.m_radius-nodeStyle.m_linewidth*2,
                        (height-y)+nodeStyle.m_radius+nodeStyle.m_linewidth*2);
            cr->line_to(x+nodeStyle.m_radius+nodeStyle.m_linewidth*2,
                        (height-y)+nodeStyle.m_radius+nodeStyle.m_linewidth*2);
            cr->line_to(x+nodeStyle.m_radius+nodeStyle.m_linewidth*2,
                        (height-y)-nodeStyle.m_radius-nodeStyle.m_linewidth*2);
            cr->close_path();
            break;

          case NODE_CIRCLE:
          default:
            cr->arc(x, height-y,
                    nodeStyle.m_radius+nodeStyle.m_linewidth*2,
                    0, 2 * M_PI);
            break;
            // end switching node shape
          }

          cr->stroke();

          // end double border
        }
        cr->restore();

        // end node had a position
      }

      // end looping over nodes
    }

    // end window exists
  }

  return true;
}
